# -*- coding: utf-8 -*-
#
# Author: GavinGong aka VisualDust
# Github: github.com/visualDust
# Date:   20231224

import os
import pathlib
import sys

from vdtoys.localstorage import get_user_app_data_directory, get_user_config_directory


def get_create_neetbox_config_directory():
    path = os.path.join(get_user_config_directory(), "neetbox")
    if not os.path.exists(path):
        os.makedirs(path)
    assert os.path.isdir(
        path
    ), f"Fialed to create neetbox config directory {path}, please check your permission or create it manually."
    return path


def get_create_neetbox_data_directory():
    path = os.path.join(get_user_app_data_directory(), "neetbox")
    if not os.path.exists(path):
        os.makedirs(path)
    assert os.path.isdir(
        path
    ), f"Fialed to create neetbox data directory {path}, please check your permission or create it manually."
    return path



import atexit
from threading import Thread

import httpx
import websocket
from vdtoys.framing import get_caller_info_traceback

from neetbox._protocol import *

httpxClient: httpx.Client = httpx.Client(  # httpx client
    proxies={
        "http://": None,
        "https://": None,
    }
)


class WebsocketClient:
    instances = {}

    def __init__(self, url, on_open, on_message, on_error, on_close, offline_message_buffer_size=0):
        self.wsApp = websocket.WebSocketApp(  # create websocket client
            url=url,
            on_open=on_open,
            on_message=on_message,
            on_error=on_error,
            on_close=on_close,
        )
        self.message_query = []  # websocket message query
        self.offline_message_buffer_size = offline_message_buffer_size
        self.creator = get_caller_info_traceback(stack_offset=1)
        WebsocketClient.instances[self.creator.strid] = self

    def connect(self, reconnect=1):
        if not self.wsApp:
            raise RuntimeError("You should setup WebsocketClient before connect.")
        Thread(
            target=self.wsApp.run_forever, kwargs={"reconnect": reconnect}, daemon=True
        ).start()  # initialize and start ws thread

    @property
    def is_connected(self) -> bool:
        return (
            self.wsApp.sock.connected
            if hasattr(self.wsApp, "sock") and self.wsApp.sock is not None
            else False
        )

    def send(self, message: EventMsg):
        self.message_query.append(message)
        if self.is_connected:  # if ws client exist
            try:
                while len(self.message_query):
                    self.wsApp.send(self.message_query[0].dumps())
                    self.message_query.pop(0)
            except Exception as e:
                pass
        else:
            while len(self.message_query) > self.offline_message_buffer_size:
                self.message_query.pop(0)


def _clean_websocket_on_exit():
    for client in WebsocketClient.instances.values():
        if hasattr(client, "wsApp") and client.wsApp:
            try:
                client.wsApp.close()
            except Exception as e:
                pass


# clean websocket connections on exit
atexit.register(_clean_websocket_on_exit)



from time import time

from transformers import (
    TrainerCallback,
    TrainerControl,
    TrainerState,
    TrainingArguments,
)

from neetbox import action, add_hyperparams, add_scalar, logger, progress
from neetbox._protocol import get_timestamp


class NeetboxTrainerCallback(TrainerCallback):
    _SCALAR_NAME_IGNORED = [  # ignore scalars that are not useful metrics
        "epoch",
        "step",
        "total_flos",
        "total_mem",
        "total_mem_cpu",
        "total_mem_gpu",
        "eval_runtime",
        "eval_samples_per_second",
        "eval_steps_per_second",
        "train_runtime",
        "train_samples_per_second",
        "train_steps_per_second",
        "train_loss",
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._status = {
            "last_update_epoch": None,
            "last_update_epoch_time": None,
            "last_update_step": None,
            "last_update_step_time": None,
        }
        self._scalars = {}
        self._launched_time_stamp = get_timestamp()

    def _extract_scalar(self, log):
        scalars = {}
        # Check if anything in last log is scalar
        for key, value in log.items():
            if key not in self._SCALAR_NAME_IGNORED and isinstance(value, (int, float)):
                scalars[key] = value
        return scalars

    def on_train_begin(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        """
        Event called at the beginning of training.
        """
        logger.info("Training started.")
        add_hyperparams(args.to_dict())

        @action(name="Trigger stop training")
        def trigger_should_training_stop():
            """Trigger TrainerControl.should_training_stop"""
            control.should_training_stop = not control.should_training_stop
            message = f"should_training_stop was set to: {control.should_training_stop}"
            logger.send_mention(message)
            return message

        @action(name="Trigger save model")
        def tirgger_should_save():
            """Trigger TrainerControl.should_save"""
            control.should_save = not control.should_save
            message = f"should_save was set to: {control.should_save}"
            logger.send_mention(message)
            return message

        @action(name="Trigger evaluate model")
        def trigger_should_evaluate():
            """Trigger TrainerControl.should_evaluate"""
            control.should_evaluate = not control.should_evaluate
            message = f"should_evaluate was set to: {control.should_evaluate}"
            logger.send_mention(message)
            return message

    def on_train_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        """
        Event called at the end of training.
        """
        logger.info("Training finished.")

    def on_save(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        """
        Event called after a checkpoint save.
        """
        log_history = state.log_history[-1] if len(state.log_history) else None
        if log_history is None:
            return  # Skip logging
        log = dict(log_history)
        scalars = self._extract_scalar(log)
        logger.info(f"Checkpoint saved on {scalars}")

    def on_step_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        """
        Event called after each step.
        """
        # Update progress bar for epoch
        if not self._status["last_update_epoch"]:
            self._status["last_update_epoch"] = state.epoch
            self._status["last_update_epoch_time"] = time()
        else:
            if state.epoch != self._status["last_update_epoch"]:
                last_update_epoch = self._status["last_update_epoch"]
                last_update_epoch_time = self._status["last_update_epoch_time"]
                elapsed_time = time() - last_update_epoch_time
                rate = (state.epoch - last_update_epoch) / elapsed_time
                progress._update(
                    name="epoch",
                    what_is_current=f"of current step {int(state.global_step)}",
                    done=int(state.epoch),
                    total=int(args.num_train_epochs),
                    rate=rate,
                    timestamp=self._launched_time_stamp,
                )
                self._status["last_update_epoch"] = state.epoch
                self._status["last_update_epoch_time"] = time()

        # Update progress bar for step
        if not self._status["last_update_step"]:
            self._status["last_update_step"] = state.global_step
            self._status["last_update_step_time"] = time()
        else:
            if state.global_step != self._status["last_update_step"]:
                last_update_step = self._status["last_update_step"]
                last_update_step_time = self._status["last_update_step_time"]
                elapsed_time = time() - last_update_step_time
                rate = (state.global_step - last_update_step) / elapsed_time
                progress._update(
                    name="step",
                    what_is_current=f"of epoch {int(state.epoch)}",
                    done=int(state.global_step),
                    total=int(state.max_steps),
                    rate=rate,
                    timestamp=self._launched_time_stamp,
                )
                self._status["last_update_step"] = state.global_step
                self._status["last_update_step_time"] = time()

    def on_log(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        """
        Event called after logging.
        """
        log_history = state.log_history[-1] if len(state.log_history) else None
        if log_history is None:
            return  # Skip logging
        log = dict(log_history)
        scalars = self._extract_scalar(log)

        for key, value in scalars.items():
            if key not in self._scalars or self._scalars[key] != value:
                add_scalar(
                    name=key,
                    x=state.global_step,
                    y=value,
                )
                self._scalars[key] = value



import argparse
import json
import sys
import subprocess
import time
from typing import Optional


def run(config):
    from .server_process import server_process

    server_process(config)


def start(config) -> Optional[int]:
    """start a server as background daemon

    Args:
        config (dict): server launch config

    Returns:
        DaemonableProcess: the daemon process
    """
    from neetbox.utils import DaemonableProcess
    import neetbox.server._daemon_server_launch_script as myself

    popen = DaemonableProcess(  # server daemon
        target=myself,
        args=["--config", json.dumps(config)],
        mode=config["mode"],
        redirect_stdout=subprocess.DEVNULL if config["mute"] else None,
        env_append={"NEETBOX_DAEMON_PROCESS": "1"},
    ).start()
    time.sleep(1)
    return popen


if __name__ == "__main__":
    print("server starting.\n name:\t", __name__, "\nargs:\t", sys.argv)
    if len(sys.argv) <= 1:
        print("Warning: empty server config")
        config = None
    else:
        ap = argparse.ArgumentParser()
        ap.add_argument("--config")
        args = ap.parse_args()
        print(type(args.config))
        print(args.config)
        config = json.loads(args.config)
        print("Daemon started with config:", config)
    run(config)
    print("server closed.")
